	.include	"jaguar.inc"

NB_STARS	equ	(9*256)

;;; operation mode for division
DIV_FRAC	equ	1

;;; to manually clear each star (double buffered)
CLEAR_STAR	equ	0
	
WIDTH	equ	320
HEIGHT	equ	240
	
RANDGEN_J	equ	3
RANDGEN_MOD	equ	3
RANDGEN_K	equ	(1<<RANDGEN_MOD)-1	; k > j (doit etre le predecesseur d'une puissance de 2)
RANDGEN_R	equ	5

	.offset
STAR_XPOS:	ds.w	1
STAR_YPOS:	ds.w	1
STAR_ZPOS:	ds.w	1
STAR_SPEED:	ds.w	1
	.if	CLEAR_STAR
STAR_OFFSET1:	ds.l	1
STAR_OFFSET2:	ds.l	1
	.endif
STAR_SIZEOF:	ds.l	0
	
	.text
	.68000
random_init:
	move.l	d2,-(sp)
	move.l	#random_seeds,a0
	moveq	#1,d1
	move.l	d1,d2
	moveq	#RANDGEN_K,d0
.init_seeds:
	eor.l	#$31415926,d1
	move.l	d1,(a0)+
	ror.l	#8,d1
	ror.l	#5,d1
	add.l	d2,d1
	dbf	d0,.init_seeds
	move.w	#RANDGEN_K*4,random_index
	move.l	(sp)+,d2
	rts

.macro	gpu_random_next
	;; \1 = r14 ou r15
	;; \2, \3 deux registres pour les calculs intermediaires
	;; \4 = pour mettre le resultat
	movei	#random_index,\1
	loadw	(\1),\2
	move	\2,\3
	addq	#(RANDGEN_K-RANDGEN_J)*4,\2
	addq	#4,\1
	shlq	#32-RANDGEN_MOD,\2
	load	(\1+\3),\4
	shrq	#30-RANDGEN_MOD,\2
	load	(\1+\2),\2
	add	\2,\4
	rorq	#RANDGEN_R,\4
	store	\4,(\1+\3)
	subq	#4,\3
	subq	#4,\1
	shlq	#32-RANDGEN_MOD,\3
	shrq	#30-RANDGEN_MOD,\3
	storew	\3,(\1)
.endm

	.phrase
starfield_routine:
	.gpu	
	.org	0
.starfield_begin:
	move	PC,r0
	movei	#_starfield_address,r20
	movei	#stars+4,r2
	movei	#511,r16	; mask
	load	(r20),r1	; screen BASE address
	addq	#4,r20
	load	(r20),r17	; CENTER_Y|CENTER_X
	addq	#4,r20
	loadw	(r20),r25	; SPEEDZ
	move	r17,r18
	shlq	#16,r17		; CENTER_X|0
	sharq	#16,r18		; CENTER_Y
	sharq	#16,r17		; CENTER_X
	movei	#1<<16,r15
	movei	#.do_one_star-.starfield_begin,r20
	movei	#.init_star-.starfield_begin,r21
	movei	#.loop_star-.starfield_begin,r22
	movei	#256,r23	; 256
	movei	#(63*16)<<16,r24; 63*16|0
	movei	#NB_STARS,r4	; i
	add	r0,r20		; relocate
	add	r0,r21		; relocate
	add	r0,r22		; relocate
	movei	#WIDTH,r5
	movei	#HEIGHT,r6
	moveq	#0,r19		; 0
.do_one_star:
	move	r15,r11		; 1<<16
	load	(r2),r7		; Z|SPEED
	.if	CLEAR_STAR
	addq	#8,r2
	move	r7,r8		; Z|SPEED
	load	(r2),r12	; OFFSET2
	subq	#4,r2
	add	r1,r12
	load	(r2),r13	; OFFSET1
	addq	#4,r2
	storeb	r19,(r12)	; clear star
	store	r13,(r2)
	subq	#12,r2
	.else
	subq	#4,r2	
	move	r7,r8		; Z|SPEED
	.endif
	sharq	#16,r7		; Z
	jump	mi,(r21)	; init star
	.if	DIV_FRAC
	shlq	#16,r7
	.else
	nop
	.endif
	jump	eq,(r21)	; init star
.draw_star:
	shlq	#16,r8		; SPEED|0
	div	r7,r11		; compute 1/z << 16 = inv_z
	.if	DIV_FRAC
	sharq	#16,r7
	.endif
	load	(r2),r9		; X|Y
	sharq	#16,r8		; SPEED
	move	r9,r10		; X|Y
	sharq	#16,r9		; X
	shlq	#16,r10		; Y|0
	move	r7,r12		; save Z
	sharq	#16,r10		; Y
	shrq	#4,r12		; index in CLUT = Z / 16
	sub	r8,r7		; Z -= SPEED	
	shlq	#16,r8		; SPEED|0
	shlq	#16,r7		; Z|0
	shrq	#16,r8		; 0|SPEED
	abs	r9
	jr	cc,.x_mult_ok
	mult	r11,r9		; x*inv_z
	neg	r9
.x_mult_ok:
	abs	r10
	jr	cc,.y_mult_ok
	mult	r11,r10		; y*inv_z
	neg	r10
.y_mult_ok:
	sharq	#6,r9		; x*inv_z>>6
	sharq	#6,r10		; y*inv_z>>6
	add	r17,r9		; x*inv_z>>6 + CENTER_X
	addqt	#4,r2		
	jr	pl,.clip_x_ge0
	or	r7,r8		; Z|SPEED
.clip_x_lt0:
	jump	(r22)		; loop_star
	store	r19,(r2)	; reinit star next time
.clip_x_ge0:
	cmp	r5,r9		; compare to WIDTH
	jr	mi,.clip_x_ltW
	nop
.clip_x_geW:
	jump	(r22)		; loop_star
	store	r19,(r2)	; reinit star next time
.clip_x_ltW:
	add	r18,r10		; y*inv_z>>6 + CENTER_Y
	jr	pl,.clip_y_ge0
	nop
.clip_y_lt0:
	jump	(r22)		; loop_star
	store	r19,(r2)	; reinit star next time
.clip_y_ge0:
	cmp	r6,r10		; compare to HEIGHT
	jr	mi,.clip_y_ltH
	mult	r5,r10
.clip_y_geH:
	jump	(r22)		; loop_star
	store	r19,(r2)	; reinit star next time
.clip_y_ltH:
	add	r10,r9
	store	r8,(r2)
	.if	CLEAR_STAR
	addq	#4,r2
	store	r9,(r2)
	.endif
	add	r1,r9		; add base address
	.if	CLEAR_STAR
	subq	#4,r2
	.endif
	jump	(r22)		; loop_star
	storeb	r12,(r9)
.init_star:
	gpu_random_next	r14,r12,r13,r11	; temp = new random value in r11
	move	r11,r9			; temp
	shrq	#9,r11			; temp >>= 9
	and	r16,r9			; temp & 511
	move	r11,r10			; temp
	sub	r23,r9			; x = temp & 511 - 256
	and	r16,r10			; temp & 511
	shrq	#9,r11			; temp >>= 9
	moveq	#7,r7
	sub	r23,r10			; y = temp & 511 - 256
	and	r11,r7			; temp & 7
	shlq	#16,r9			; X|0
	shlq	#16,r10			; Y|0
	shrq	#16,r9			; 0|X
	add	r25,r7			; speed = temp & 7 + speedz
	or	r9,r10
	shlq	#16,r7			; SPEED|0
	store	r10,(r2)
	shrq	#16,r7			; 0|SPEED
	addqt	#4,r2
	or	r24,r7			; Z|SPEED
	store	r7,(r2)
.loop_star:
	subq	#1,r4		; i--
	jump	ne,(r20)
	addq	#STAR_SIZEOF,r2
.starfield_return:
	movei	#starfield_done,r0
	moveq	#1,r1
	storew	r1,(r0)
	load	(r31),r0
	jump	(r0)
	addq	#4,r31
	.long
.starfield_end:

STARFIELD_ROUTINE_SIZE	equ	.starfield_end-.starfield_begin

	.print	"GPU Routine size = ",STARFIELD_ROUTINE_SIZE
	
	.68000

        .extern _bcopy
        .globl  _init_starfield_routine

_init_starfield_routine:	
        pea     STARFIELD_ROUTINE_SIZE
        move.l  4+4(sp),-(sp)
        pea     starfield_routine
        jsr     _bcopy
        lea     12(sp),sp

	bsr	random_init
	move.l	#stars,a0
	move.w	#NB_STARS-1,d0
.clear:
	clr.w	STAR_ZPOS(a0)
	.if	CLEAR_STAR
	clr.l	STAR_OFFSET1(a0)
	clr.l	STAR_OFFSET2(a0)	
	.endif
	lea	STAR_SIZEOF(a0),a0
	dbf	d0,.clear
	move.l	#_starfield_center_y,a0
	move.w	#HEIGHT/2,(a0)+	; centery
	move.w	#WIDTH/2,(a0)+	; centerx
	move.w	#8,(a0)+	; speedz
	
        move.l  #STARFIELD_ROUTINE_SIZE,d0
        rts

	.globl	_go_starfield
	.extern	_jump_gpu_subroutine
_go_starfield:
	clr.w	starfield_done
	jmp	_jump_gpu_subroutine

	.globl	_wait_starfield
_wait_starfield:
	tst.w	starfield_done
	beq.s	_wait_starfield
	rts
	
	.bss
	.long
random_index:
	ds.w	1
	ds.w	1
random_seeds:
	ds.l	RANDGEN_K+1

	.bss
	.even
starfield_done:
	ds.w	1

	.bss
	.globl	_starfield_address
	.globl	_starfield_center_y
	.globl	_starfield_center_x
	.globl	_starfield_speedz	
	.long
_starfield_address:	ds.l	1
_starfield_center_y:	ds.w	1
_starfield_center_x:	ds.w	1	
_starfield_speedz:	ds.w	1
	
	.bss
	.long
stars:
	.rept	NB_STARS
	ds.b	STAR_SIZEOF
	.endr

